@@ -4,7 +4,7 @@
 /* Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2001,2005 James Aylett
  * Copyright 2001,2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016,2017 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018 Olly Betts
  * Copyright 2009 Frank J Bruzzaniti
  * Copyright 2012 Mihai Bivol
  *
@@ -37,11 +37,11 @@
 
 #include <sys/types.h>
 #include "safeunistd.h"
+#include <cerrno>
 #include <cstdio>
 #include <cstdlib>
 #include <cstring>
 #include "safefcntl.h"
-#include "safeerrno.h"
 #include <ctime>
 
 #include <xapian.h>
@@ -178,6 +178,14 @@ index_add_default_filters()
     index_command("application/vnd.ms-outlook",
 		  Filter(get_pkglibbindir() + "/outlookmsg2html", "text/html",
 			 false));
+    index_command("application/vnd.ms-visio.drawing",
+		  Filter("vsd2xhtml", "image/svg+xml", false));
+    index_command("application/vnd.ms-visio.stencil",
+		  Filter("vsd2xhtml", "image/svg+xml", false));
+    index_command("application/vnd.ms-visio.template",
+		  Filter("vsd2xhtml", "image/svg+xml", false));
+    index_command("application/vnd.visio",
+		  Filter("vsd2xhtml", "image/svg+xml", false));
     // pod2text's output character set doesn't seem to be documented, but from
     // inspecting the source it looks like it's probably iso-8859-1.  We need
     // to pass "--errors=stderr" or else minor POD formatting errors cause a
@@ -207,6 +215,12 @@ index_add_default_filters()
 			 false));
     index_command("text/vcard",
 		  Filter(get_pkglibbindir() + "/vcard2text", false));
+    index_command("application/vnd.apply.keynote",
+		  Filter("key2text", false));
+    index_command("application/vnd.apply.numbers",
+		  Filter("numbers2text", false));
+    index_command("application/vnd.apply.pages",
+		  Filter("pages2text", false));
 }
 
 void
@@ -235,19 +249,23 @@ index_init(const string & dbpath, const Xapian::Stem & stemmer,
     if (!overwrite) {
 	db = Xapian::WritableDatabase(dbpath, Xapian::DB_CREATE_OR_OPEN);
 	old_docs_not_seen = db.get_doccount();
-	old_lastdocid = db.get_lastdocid();
-	if (delete_removed_documents) {
-	    // + 1 so that old_lastdocid is a valid subscript.
-	    updated.resize(old_lastdocid + 1);
-	}
-	try {
-	    Xapian::valueno slot = use_ctime ? VALUE_CTIME : VALUE_LASTMOD;
-	    string ubound = db.get_value_upper_bound(slot);
-	    if (!ubound.empty())
-		last_altered_max = binary_string_to_int(ubound);
-	} catch (const Xapian::UnimplementedError &) {
-	    numeric_limits<time_t> n;
-	    last_altered_max = n.max();
+	// Handle an initially empty database exactly the same way as when
+	// overwrite is true.
+	if (old_docs_not_seen != 0) {
+	    old_lastdocid = db.get_lastdocid();
+	    if (delete_removed_documents) {
+		// + 1 so that old_lastdocid is a valid subscript.
+		updated.resize(old_lastdocid + 1);
+	    }
+	    try {
+		Xapian::valueno slot = use_ctime ? VALUE_CTIME : VALUE_LASTMOD;
+		string ubound = db.get_value_upper_bound(slot);
+		if (!ubound.empty())
+		    last_altered_max = binary_string_to_int(ubound);
+	    } catch (const Xapian::UnimplementedError &) {
+		numeric_limits<time_t> n;
+		last_altered_max = n.max();
+	    }
 	}
     } else {
 	db = Xapian::WritableDatabase(dbpath, Xapian::DB_CREATE_OR_OVERWRITE);
@@ -296,7 +314,7 @@ parse_pdfinfo_field(const char * p, const char * end, string & out, const char *
 
 static void
 get_pdf_metainfo(const string & file, string &author, string &title,
-		 string &keywords, string &topic)
+		 string &keywords, string &topic, int& pages)
 {
     try {
 	string cmd = "pdfinfo -enc UTF-8";
@@ -322,6 +340,13 @@ get_pdf_metainfo(const string & file, string &author, string &title,
 		case 'K':
 		    PARSE_PDFINFO_FIELD(start, eol, keywords, "Keywords");
 		    break;
+		case 'P': {
+		    string s;
+		    PARSE_PDFINFO_FIELD(start, eol, s, "Pages");
+		    if (!s.empty())
+			pages = atoi(s.c_str());
+		    break;
+		}
 		case 'S':
 		    PARSE_PDFINFO_FIELD(start, eol, topic, "Subject");
 		    break;
@@ -444,6 +469,12 @@ index_check_existing(const string & urlterm, time_t last_altered,
     return false;
 }
 
+void
+index_remove_failed_entry(const string& urlterm)
+{
+    failed.del(urlterm);
+}
+
 void
 index_add_document(const string & urlterm, time_t last_altered,
 		   Xapian::docid did, const Xapian::Document & doc)
@@ -518,6 +549,7 @@ index_mimetype(const string & file, const string & urlterm, const string & url,
     string author, title, sample, keywords, topic, dump;
     string md5;
     time_t created = time_t(-1);
+    int pages = -1;
 
     map<string, Filter>::const_iterator cmd_it = commands.find(mimetype);
     if (cmd_it == commands.end()) {
@@ -586,6 +618,8 @@ index_mimetype(const string & file, const string & urlterm, const string & url,
 			    // error messages from the command.
 			    if (cmd_it->second.output_type == "text/html") {
 				tmpout = get_tmpfile("tmp.html");
+			    } else if (cmd_it->second.output_type == "image/svg+xml") {
+				tmpout = get_tmpfile("tmp.svg");
 			    } else {
 				tmpout = get_tmpfile("tmp.txt");
 			    }
@@ -616,7 +650,7 @@ index_mimetype(const string & file, const string & urlterm, const string & url,
 		if (!tmpout.empty()) {
 		    // Output in temporary file.
 		    (void)stdout_to_string(cmd, use_shell);
-		    if (!load_file(tmpout, dump)) {
+		    if (!load_file(tmpout, dump, NOCACHE)) {
 			throw ReadError("Couldn't read output file");
 		    }
 		    unlink(tmpout.c_str());
@@ -651,6 +685,14 @@ index_mimetype(const string & file, const string & urlterm, const string & url,
 		    sample = p.sample;
 		    author = p.author;
 		    created = p.created;
+		} else if (cmd_it->second.output_type == "image/svg+xml") {
+		    SvgParser svgparser;
+		    svgparser.parse(dump);
+		    dump = svgparser.dump;
+		    title = svgparser.title;
+		    keywords = svgparser.keywords;
+		    // FIXME: topic = svgparser.topic;
+		    author = svgparser.author;
 		} else if (!charset.empty()) {
 		    convert_to_utf8(dump, charset);
 		}
@@ -717,7 +759,7 @@ index_mimetype(const string & file, const string & urlterm, const string & url,
 				d.get_size(), d.get_mtime());
 		return;
 	    }
-	    get_pdf_metainfo(file, author, title, keywords, topic);
+	    get_pdf_metainfo(file, author, title, keywords, topic, pages);
 	} else if (mimetype == "application/postscript") {
 	    // There simply doesn't seem to be a Unicode capable PostScript to
 	    // text converter (e.g. pstotext always outputs ISO-8859-1).  The
@@ -755,7 +797,7 @@ index_mimetype(const string & file, const string & urlterm, const string & url,
 		throw;
 	    }
 	    try {
-		get_pdf_metainfo(tmpfile, author, title, keywords, topic);
+		get_pdf_metainfo(tmpfile, author, title, keywords, topic, pages);
 	    } catch (...) {
 		unlink(tmpfile.c_str());
 		throw;
@@ -970,7 +1012,7 @@ index_mimetype(const string & file, const string & urlterm, const string & url,
 	}
 
 	// Compute the MD5 of the file if we haven't already.
-	if (md5.empty() && md5_file(file, md5, d.try_noatime()) == 0) {
+	if (md5.empty() && !d.md5(md5)) {
 	    if (errno == ENOENT || errno == ENOTDIR) {
 		skip(urlterm, context, "File removed during indexing",
 		     d.get_size(), d.get_mtime(),
@@ -1042,6 +1084,10 @@ index_mimetype(const string & file, const string & urlterm, const string & url,
 	    record += "\ncreated=";
 	    record += str(created);
 	}
+	if (pages >= 0) {
+	    record += "\npages=";
+	    record += str(pages);
+	}
 	off_t size = d.get_size();
 	record += "\nsize=";
 	record += str(size);
@@ -1177,6 +1223,14 @@ index_mimetype(const string & file, const string & urlterm, const string & url,
 	     SKIP_VERBOSE_ONLY | SKIP_SHOW_FILENAME);
     } catch (const std::string & error) {
 	skip(urlterm, context, error, d.get_size(), d.get_mtime());
+    } catch (const std::bad_alloc&) {
+	// Attempt to flag the file as failed and commit changes, though that
+	// might fail too if we're low on memory rather than being asked to
+	// allocate a ludicrous amount.
+	skip(urlterm, context, "Out of memory trying to extract text from file",
+	     d.get_size(), d.get_mtime(),
+	     SKIP_SHOW_FILENAME);
+	throw CommitAndExit("Caught std::bad_alloc", "");
     }
 }
 
