@@ -3,7 +3,7 @@
  * Copyright 1999,2000,2001 BrightStation PLC
  * Copyright 2001,2005 James Aylett
  * Copyright 2001,2002 Ananova Ltd
- * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2017 Olly Betts
+ * Copyright 2002,2003,2004,2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2017,2018 Olly Betts
  * Copyright 2009 Frank J Bruzzaniti
  * Copyright 2012 Mihai Bivol
  *
@@ -26,6 +26,7 @@
 #include <config.h>
 
 #include <algorithm>
+#include <fstream>
 #include <iostream>
 #include <string>
 #include <map>
@@ -37,6 +38,10 @@
 #include <cstring>
 #include "safefcntl.h"
 
+#ifdef HAVE_FNMATCH
+# include <fnmatch.h>
+#endif
+
 #include <xapian.h>
 
 #include "commonhelp.h"
@@ -68,7 +73,11 @@ static double sleep_before_opendir = 0;
 static string root;
 static string url_start_path;
 
-inline static bool
+#ifdef HAVE_FNMATCH
+static vector<pair<const char*, const char*>> mime_patterns;
+#endif
+
+static inline bool
 p_notalnum(unsigned int c)
 {
     return !C_isalnum(c);
@@ -78,30 +87,58 @@ static void
 index_file(const string &file, const string &url, DirectoryIterator & d,
 	   map<string, string>& mime_map)
 {
-    string ext;
-    const char * dot_ptr = strrchr(d.leafname(), '.');
-    if (dot_ptr) {
-	ext.assign(dot_ptr + 1);
-	if (ext.size() > max_ext_len)
-	    ext.resize(0);
-    }
-
     string urlterm("U");
     urlterm += url;
 
     if (urlterm.length() > MAX_SAFE_TERM_LENGTH)
 	urlterm = hash_long_term(urlterm, MAX_SAFE_TERM_LENGTH);
 
-    string mimetype = mimetype_from_ext(mime_map, ext);
-    if (mimetype == "ignore") {
-	return;
-    } else if (mimetype == "skip") {
-	// Ignore mimetype, skipped mimetype should not be quietly ignored.
-	string m = "skipping extension '";
-	m += ext;
-	m += "'";
-	skip(urlterm, file.substr(root.size()), m, d.get_size(), d.get_mtime());
-	return;
+    const char* leafname = d.leafname();
+
+    string mimetype;
+#ifdef HAVE_FNMATCH
+    for (auto&& i : mime_patterns) {
+	if (fnmatch(i.first, leafname, 0) == 0) {
+	    if (strcmp(i.second, "ignore") == 0)
+		return;
+	    if (strcmp(i.second, "skip") == 0) {
+		string m = "Leafname '";
+		m += leafname;
+		m += "' matches pattern: ";
+		m += i.first;
+		skip(urlterm, file.substr(root.size()), m,
+		     d.get_size(), d.get_mtime());
+		return;
+	    }
+	    mimetype = i.second;
+	    break;
+	}
+    }
+#endif
+
+    string ext;
+    const char * dot_ptr = strrchr(leafname, '.');
+    if (dot_ptr) {
+	ext.assign(dot_ptr + 1);
+	if (ext.size() > max_ext_len)
+	    ext.resize(0);
+    }
+
+    if (mimetype.empty()) {
+	mimetype = mimetype_from_ext(mime_map, ext);
+	if (mimetype == "ignore") {
+	    // Remove any existing failed entry for this file.
+	    index_remove_failed_entry(urlterm);
+	    return;
+	} else if (mimetype == "skip") {
+	    // Ignore mimetype, skipped mimetype should not be quietly ignored.
+	    string m = "skipping extension '";
+	    m += ext;
+	    m += "'";
+	    skip(urlterm, file.substr(root.size()), m,
+		 d.get_size(), d.get_mtime());
+	    return;
+	}
     }
 
     // Check the file size.
@@ -247,6 +284,63 @@ parse_size(char* p)
     return -1;
 }
 
+static bool
+parse_filter_rule(const char* rule, map<string, string>& mime_map)
+{
+    const char* s = strchr(rule, ':');
+    if (s == NULL || s[1] == '\0') {
+	cerr << "Invalid filter mapping '" << rule << "'\n"
+		"Should be of the form TYPE:COMMAND or TYPE1,TYPE2:COMMAND or "
+		"TYPE,EXT:COMMAND\n"
+		"e.g. 'application/octet-stream:strings -n8'"
+	     << endl;
+	return false;
+    }
+
+    const char* c = static_cast<const char*>(memchr(rule, ',', s - rule));
+    string output_type, output_charset;
+    if (c) {
+	// Filter produces a specified content-type.
+	++c;
+	const char* c2 = static_cast<const char *>(memchr(c, ',', s - c));
+	if (c2) {
+	    output_type.assign(c, c2 - c);
+	    ++c2;
+	    output_charset.assign(c2, s - c2);
+	} else {
+	    output_type.assign(c, s - c);
+	}
+	--c;
+	if (output_type.find('/') == string::npos) {
+	    auto m = mime_map.find(output_type);
+	    if (m != mime_map.end()) {
+		output_type = m->second;
+	    } else {
+		const char* r = built_in_mime_map(output_type);
+		if (r) output_type = r;
+	    }
+	}
+	if (output_type != "text/html" &&
+	    output_type != "text/plain" &&
+	    output_type != "image/svg+xml") {
+	    cerr << "Currently only output types 'image/svg+xml', "
+		    "'text/html' and 'text/plain' are supported."
+		 << endl;
+	    return false;
+	}
+    } else {
+	c = s;
+    }
+
+    const char* cmd = s + 1;
+    // Analyse the command string to decide if it needs a shell.
+    bool use_shell = command_needs_shell(cmd);
+    index_command(string(rule, c - rule),
+		  Filter(string(cmd), output_type, output_charset, use_shell));
+
+    return true;
+}
+
 int
 main(int argc, char **argv)
 {
@@ -270,32 +364,40 @@ main(int argc, char **argv)
     string site_term, host_term;
     Xapian::Stem stemmer("english");
 
-    enum { OPT_OPENDIR_SLEEP = 256, OPT_SAMPLE };
+    enum {
+	OPT_OPENDIR_SLEEP = 256,
+	OPT_SAMPLE,
+	OPT_READ_FILTERS
+    };
+    constexpr auto NO_ARG = no_argument;
+    constexpr auto REQ_ARG = required_argument;
     static const struct option longopts[] = {
-	{ "help",	no_argument,		NULL, 'h' },
-	{ "version",	no_argument,		NULL, 'V' },
-	{ "overwrite",	no_argument,		NULL, 'o' },
-	{ "duplicates",	required_argument,	NULL, 'd' },
-	{ "no-delete",	no_argument,		NULL, 'p' },
-	{ "preserve-nonduplicates",	no_argument,	NULL, 'p' },
-	{ "db",		required_argument,	NULL, 'D' },
-	{ "url",	required_argument,	NULL, 'U' },
-	{ "mime-type",	required_argument,	NULL, 'M' },
-	{ "filter",	required_argument,	NULL, 'F' },
-	{ "depth-limit",required_argument,	NULL, 'l' },
-	{ "follow",	no_argument,		NULL, 'f' },
-	{ "ignore-exclusions",	no_argument,	NULL, 'i' },
-	{ "stemmer",	required_argument,	NULL, 's' },
-	{ "spelling",	no_argument,		NULL, 'S' },
-	{ "verbose",	no_argument,		NULL, 'v' },
-	{ "empty-docs",	required_argument,	NULL, 'e' },
-	{ "max-size",	required_argument,	NULL, 'm' },
-	{ "sample",	required_argument,	NULL, OPT_SAMPLE },
-	{ "sample-size",required_argument,	NULL, 'E' },
-	{ "title-size",	required_argument,	NULL, 'T' },
-	{ "retry-failed",	no_argument,	NULL, 'R' },
-	{ "opendir-sleep",	required_argument,	NULL, OPT_OPENDIR_SLEEP },
-	{ "track-ctime",no_argument,		NULL, 'C' },
+	{ "help",		NO_ARG,		NULL, 'h' },
+	{ "version",		NO_ARG,		NULL, 'V' },
+	{ "overwrite",		NO_ARG,		NULL, 'o' },
+	{ "duplicates",		REQ_ARG,	NULL, 'd' },
+	{ "no-delete",		NO_ARG,		NULL, 'p' },
+	{ "preserve-nonduplicates", NO_ARG,	NULL, 'p' },
+	{ "db",			REQ_ARG,	NULL, 'D' },
+	{ "url",		REQ_ARG,	NULL, 'U' },
+	{ "mime-type",		REQ_ARG,	NULL, 'M' },
+	{ "mime-type-match",	REQ_ARG,	NULL, 'G' },
+	{ "filter",		REQ_ARG,	NULL, 'F' },
+	{ "read-filters",	REQ_ARG,	NULL, OPT_READ_FILTERS },
+	{ "depth-limit",	REQ_ARG,	NULL, 'l' },
+	{ "follow",		NO_ARG,		NULL, 'f' },
+	{ "ignore-exclusions",	NO_ARG,		NULL, 'i' },
+	{ "stemmer",		REQ_ARG,	NULL, 's' },
+	{ "spelling",		NO_ARG,		NULL, 'S' },
+	{ "verbose",		NO_ARG,		NULL, 'v' },
+	{ "empty-docs",		REQ_ARG,	NULL, 'e' },
+	{ "max-size",		REQ_ARG,	NULL, 'm' },
+	{ "sample",		REQ_ARG,	NULL, OPT_SAMPLE },
+	{ "sample-size",	REQ_ARG,	NULL, 'E' },
+	{ "title-size",		REQ_ARG,	NULL, 'T' },
+	{ "retry-failed",	NO_ARG,		NULL, 'R' },
+	{ "opendir-sleep",	REQ_ARG,	NULL, OPT_OPENDIR_SLEEP },
+	{ "track-ctime",	NO_ARG,		NULL, 'C' },
 	{ 0, 0, NULL, 0 }
     };
 
@@ -336,15 +438,24 @@ main(int argc, char **argv)
 "  -U, --url=URL             base url BASEDIR corresponds to (default: /)\n"
 "  -M, --mime-type=EXT:TYPE  assume any file with extension EXT has MIME\n"
 "                            Content-Type TYPE, instead of using libmagic\n"
-"                            (empty TYPE removes any existing mapping for EXT)\n"
+"                            (empty TYPE removes any existing mapping for EXT;\n"
+"                            other special TYPE values: 'ignore' and 'skip')\n"
+"  -G, --mime-type-match=GLOB:TYPE\n"
+"                            assume any file with leaf name matching shell\n"
+"                            wildcard pattern GLOB has MIME Content-Type TYPE\n"
+"                            (special TYPE values: 'ignore' and 'skip')\n"
 "  -F, --filter=M[,[T][,C]]:CMD\n"
 "                            process files with MIME Content-Type M using\n"
 "                            command CMD, which produces output (on stdout or\n"
 "                            in a temporary file) with format T (Content-Type\n"
-"                            or file extension; currently txt (default) or\n"
-"                            html) in character encoding C (default: UTF-8).\n"
+"                            or file extension; currently txt (default), html\n"
+"                            or svg) in character encoding C (default: UTF-8).\n"
 "                            E.g. -Fapplication/octet-stream:'strings -n8'\n"
 "                            or -Ftext/x-foo,,utf-16:'foo2utf16 %f %t'\n"
+"      --read-filters=FILE   bulk-load --filter arguments from FILE, which\n"
+"                            should contain one such argument per line (e.g.\n"
+"                            text/x-bar:bar2txt --utf8).  Lines starting with #\n"
+"                            are treated as comments and ignored.\n"
 "  -l, --depth-limit=LIMIT   set recursion limit (0 = unlimited)\n"
 "  -f, --follow              follow symbolic links\n"
 "  -i, --ignore-exclusions   ignore meta robots tags and similar exclusions\n"
@@ -415,10 +526,10 @@ main(int argc, char **argv)
 	    follow_symlinks = true;
 	    break;
 	case 'M': {
-	    const char * s = strchr(optarg, ':');
+	    const char * s = strrchr(optarg, ':');
 	    if (s == NULL) {
 		cerr << "Invalid MIME mapping '" << optarg << "'\n"
-			"Should be of the form ext:type, e.g. txt:text/plain\n"
+			"Should be of the form EXT:TYPE, e.g. txt:text/plain\n"
 			"(or txt: to delete a default mapping)" << endl;
 		return 1;
 	    }
@@ -429,58 +540,26 @@ main(int argc, char **argv)
 	    max_ext_len = max(max_ext_len, strlen(s + 1));
 	    break;
 	}
-	case 'F': {
-	    const char * s = strchr(optarg, ':');
-	    if (s != NULL && s[1]) {
-		const char * c =
-		    static_cast<const char *>(memchr(optarg, ',', s - optarg));
-		string output_type, output_charset;
-		if (c) {
-		    // Filter produces a specified content-type.
-		    ++c;
-		    const char * c2 =
-			static_cast<const char *>(memchr(c, ',', s - c));
-		    if (c2) {
-			output_type.assign(c, c2 - c);
-			++c2;
-			output_charset.assign(c2, s - c2);
-		    } else {
-			output_type.assign(c, s - c);
-		    }
-		    --c;
-		    if (output_type.find('/') == string::npos) {
-			map<string, string>::const_iterator m;
-			m = mime_map.find(output_type);
-			if (m != mime_map.end()) {
-			    output_type = m->second;
-			} else {
-			    const char * r = built_in_mime_map(output_type);
-			    if (r) output_type = r;
-			}
-		    }
-		    if (output_type != "text/html" &&
-			output_type != "text/plain") {
-			cerr << "Currently only output types 'text/html' and 'text/plain' are supported."
-			     << endl;
-			return 1;
-		    }
-		} else {
-		    c = s;
-		}
-
-		const char * cmd = s + 1;
-		// Analyse the command string to decide if it needs a shell.
-		bool use_shell = command_needs_shell(cmd);
-		index_command(string(optarg, c - optarg),
-			      Filter(string(cmd), output_type,
-				     output_charset, use_shell));
-	    } else {
-		cerr << "Invalid filter mapping '" << optarg << "'\n"
-			"Should be of the form TYPE:COMMAND or TYPE1,TYPE2:COMMAND or TYPE,EXT:COMMAND\n"
-			"e.g. 'application/octet-stream:strings -n8'"
-		     << endl;
+	case 'F':
+	    if (!parse_filter_rule(optarg, mime_map))
+		return 1;
+	    break;
+	case OPT_READ_FILTERS: {
+	    ifstream stream(optarg);
+	    if (!stream) {
+		cerr << "Unable to open filter file '" << optarg << "' "
+			"(" << strerror(errno) << ')' << endl;
 		return 1;
 	    }
+	    string rule;
+	    bool all_ok = true;
+	    while (getline(stream, rule)) {
+		if (startswith(rule, '#')) continue;
+		if (!parse_filter_rule(rule.c_str(), mime_map))
+		    all_ok = false;
+	    }
+	    if (!all_ok)
+		return 1;
 	    break;
 	}
 	case 'D':
@@ -584,6 +663,42 @@ main(int argc, char **argv)
 	case 'C':
 	    use_ctime = true;
 	    break;
+	case 'G': {
+	    char * s = strrchr(optarg, ':');
+	    if (s == NULL) {
+		cerr << "Invalid MIME mapping '" << optarg << "'\n"
+			"Should be of the form GLOB:TYPE, e.g. *~:ignore"
+		     << endl;
+		return 1;
+	    }
+#ifndef HAVE_FNMATCH
+	    cerr << "--mime-type-match isn't supported in this build because "
+		    "the fnmatch() function wasn't found at configure time."
+		 << endl;
+	    return 1;
+#else
+	    if (s == optarg) {
+		cerr << "--mime-type-match with an empty pattern can never "
+			"match." << endl;
+		return 1;
+	    }
+	    if (memchr(optarg, '/', s - optarg)) {
+		cerr << "--mime-type-match only matches against the leaf "
+			"filename so a pattern containing '/' can never match."
+		     << endl;
+		return 1;
+	    }
+	    const char* type = s + 1;
+	    if (*type == '\0') {
+		cerr << "--mime-type-match doesn't support an empty MIME type"
+		     << endl;
+		return 1;
+	    }
+	    *s = '\0';
+	    mime_patterns.emplace_back(optarg, type);
+	    break;
+#endif
+	}
 	case ':': // missing param
 	    return 1;
 	case '?': // unknown option: FIXME -> char
